
#cython: language_level=3str

# THIS IS AUTO GENERATED FILE, DO NOT MODIFY THIS FILE
# USE `autogen_pyd.py` FOR UPDATING

cimport numpy as np

ctypedef (double, double) tuple_double2
ctypedef (double, double, double) tuple_double3
ctypedef (double, double, double, double) tuple_double4

ctypedef fused numeric_dtype:
    int
    double
    long long



#############################################
# START FUNCTIONS COPY FROM '_bool_func.pxi'

cdef inline v(double[:] line, int offset):
    """V型拐点"""
    cdef double a, b, c
    a, b, c = line[offset - 2], line[offset - 1], line[offset]
    return b <= a and b < c


cdef inline vdown(double[:] line, int offset):
    """/\型拐点"""
    cdef double a, b, c
    a, b, c = line[offset - 2], line[offset - 1], line[offset]
    return b >= a and b > c


cdef inline kup(double[:] line, int offset):
    """斜率向上"""
    return line[offset] > line[offset - 1]


cdef inline kdown(double[:] line, int offset):
    """斜率向下"""
    return line[offset] < line[offset - 1]


cdef inline ne_l(double[:] line1, double[:] line2, int offset):
    """line1[offset] != line2[offset]"""
    return line1[offset] != line2[offset]


cdef inline eq_l(double[:] line1, double[:] line2, int offset):
    """line1[offset] == line2[offset]"""
    return line1[offset] == line2[offset]


cdef inline lt_l(double[:] line1, double[:] line2, int offset):
    """line1[offset] < line2[offset]"""
    return line1[offset] < line2[offset]


cdef inline gt_l(double[:] line1, double[:] line2):
    """line1[offset] > line2[offset]"""
    return line1[-1] > line2[-1]


cdef inline lte_l(double[:] line1, double[:] line2, int offset):
    """line1[offset] <= line2[offset]"""
    return line1[offset] <= line2[offset]


cdef inline gte_l(double[:] line1, double[:] line2, int offset):
    """line1[offset] >= line2[offset]"""
    return line1[offset] >= line2[offset]


cdef inline ne(double[:] line, double num, int offset):
    """line[offset] != num"""
    return line[offset] != num


cdef inline eq(double[:] line, double num, int offset):
    """line[offset] == num"""
    return line[offset] == num


cdef inline lt(double[:] line, double num, int offset):
    """line[offset] < num"""
    return line[offset] < num


cdef inline gt(double[:] line, double num, int offset):
    """line[offset] > num"""
    return line[offset] > num

cdef inline lte(double[:] line, double num, int offset):
    """line[offset] <= num"""
    return line[offset] <= num


cdef inline gte(double[:] line, double num, int offset):
    """line[offset] >= num"""
    return line[offset] >= num


cdef inline crossup(double[:] line1, double[:] line2, int offset):
    """金叉: line1 上穿 line2"""
    cdef int pre_offset = offset - 1
    return (line1[offset] > line2[offset]) and (line1[pre_offset] <= line2[pre_offset])


cdef inline crossdown(double[:] line1, double[:] line2, int offset):
    """死叉: line1 下穿 line2"""
    cdef int pre_offset = offset - 1
    return (line1[offset] < line2[offset]) and (line1[pre_offset] >= line2[pre_offset])


cdef inline crossup_value(double[:] line, double value, int offset):
    """上穿value"""
    return (line[offset] > value) and (line[offset - 1] <= value)


cdef inline crossdown_value(double[:] line, double value, int offset):
    """下穿value"""
    return (line[offset] < value) and (line[offset - 1] >= value)


cdef inline openup(double[:] line1, double[:] line2, int offset):
    """开口增大"""
    cdef:
        int pre_offset
        double n1, n2, m1, m2
    pre_offset = offset - 1
    n1, n2 = line1[offset], line1[pre_offset]
    m1, m2 = line2[offset], line2[pre_offset]
    if n1 > m1 and n2 > m2:
        return (n1 - m1) > (n2 - m2)
    elif n1 < m1 and n2 < m2:
        return (m1 - n1) > (m2 - n2)
    else:
        return False


cdef inline opendown(double[:] line1, double[:] line2, int offset):
    """开口缩小"""
    cdef:
        int pre_offset
        double n1, n2, m1, m2
    pre_offset = offset - 1
    n1, n2 = line1[offset], line1[pre_offset]
    m1, m2 = line2[offset], line2[pre_offset]
    if n1 > m1 and n2 > m2:
        return (n1 - m1) < (n2 - m2)
    elif n1 < m1 and n2 < m2:
        return (m1 - n1) < (m2 - n2)
    else:
        return False


cdef inline red_bar(double[:] OPEN, double[:] CLOSE, int offset):
    """ 阳线，如果是十字星并且今收>昨收，也当成阳线 """
    cdef double o, c, pre_c
    o, c, pre_c = OPEN[offset], CLOSE[offset], CLOSE[offset - 1]
    return c > o or (c == o and c > pre_c)


cdef inline red_bar_real(double[:] OPEN, double[:] CLOSE, int offset):
    """ 上涨线 """
    cdef double o, c, pre_c
    o, c, pre_c = OPEN[offset], CLOSE[offset], CLOSE[offset - 1]
    return c > pre_c or (c == pre_c and c > o)


cdef inline green_bar(double[:] OPEN, double[:] CLOSE, int offset):
    """ 阴线，如果是十字星并且今收<=昨收，也当成阴线 """
    cdef double o, c, pre_c
    o, c, pre_c = OPEN[offset], CLOSE[offset], CLOSE[offset - 1]
    return c < o or (c == o and c <= pre_c)


cdef inline green_bar_real(double[:] OPEN, double[:] CLOSE, int offset):
    """ 下跌线 """
    cdef double o, c, pre_c
    o, c, pre_c = OPEN[offset], CLOSE[offset], CLOSE[offset - 1]
    return c < pre_c or (c == pre_c and c <= o)

# END FUNCTIONS COPY FROM '_bool_func.pxi'
#############################################




#############################################
# START FUNCTIONS GEN FROM '_ta_lib_stream.pxi'

cdef stream_ACOS( np.ndarray real )
cdef stream_AD( np.ndarray high , np.ndarray low , np.ndarray close , np.ndarray volume )
cdef stream_ADD( np.ndarray real0 , np.ndarray real1 )
cdef stream_ADOSC( np.ndarray high , np.ndarray low , np.ndarray close , np.ndarray volume , int fastperiod, int slowperiod)
cdef stream_ADX( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef stream_ADXR( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef stream_APO( np.ndarray real , int fastperiod, int slowperiod, int matype)
cdef stream_AROON( np.ndarray high , np.ndarray low , int timeperiod)
cdef stream_AROONOSC( np.ndarray high , np.ndarray low , int timeperiod)
cdef stream_ASIN( np.ndarray real )
cdef stream_ATAN( np.ndarray real )
cdef stream_ATR( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef stream_AVGPRICE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_BBANDS( np.ndarray real , int timeperiod, double nbdevup, double nbdevdn, int matype)
cdef stream_BETA( np.ndarray real0 , np.ndarray real1 , int timeperiod)
cdef stream_BOP( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CCI( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef stream_CDL2CROWS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDL3BLACKCROWS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDL3INSIDE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDL3LINESTRIKE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDL3OUTSIDE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDL3STARSINSOUTH( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDL3WHITESOLDIERS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLABANDONEDBABY( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef stream_CDLADVANCEBLOCK( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLBELTHOLD( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLBREAKAWAY( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLCLOSINGMARUBOZU( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLCONCEALBABYSWALL( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLCOUNTERATTACK( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLDARKCLOUDCOVER( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef stream_CDLDOJI( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLDOJISTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLDRAGONFLYDOJI( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLENGULFING( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLEVENINGDOJISTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef stream_CDLEVENINGSTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef stream_CDLGAPSIDESIDEWHITE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLGRAVESTONEDOJI( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLHAMMER( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLHANGINGMAN( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLHARAMI( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLHARAMICROSS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLHIGHWAVE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLHIKKAKE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLHIKKAKEMOD( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLHOMINGPIGEON( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLIDENTICAL3CROWS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLINNECK( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLINVERTEDHAMMER( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLKICKING( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLKICKINGBYLENGTH( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLLADDERBOTTOM( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLLONGLEGGEDDOJI( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLLONGLINE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLMARUBOZU( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLMATCHINGLOW( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLMATHOLD( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef stream_CDLMORNINGDOJISTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef stream_CDLMORNINGSTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef stream_CDLONNECK( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLPIERCING( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLRICKSHAWMAN( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLRISEFALL3METHODS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLSEPARATINGLINES( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLSHOOTINGSTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLSHORTLINE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLSPINNINGTOP( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLSTALLEDPATTERN( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLSTICKSANDWICH( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLTAKURI( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLTASUKIGAP( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLTHRUSTING( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLTRISTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLUNIQUE3RIVER( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLUPSIDEGAP2CROWS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CDLXSIDEGAP3METHODS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_CEIL( np.ndarray real )
cdef stream_CMO( np.ndarray real , int timeperiod)
cdef stream_CORREL( np.ndarray real0 , np.ndarray real1 , int timeperiod)
cdef stream_COS( np.ndarray real )
cdef stream_COSH( np.ndarray real )
cdef stream_DEMA( np.ndarray real , int timeperiod)
cdef stream_DIV( np.ndarray real0 , np.ndarray real1 )
cdef stream_DX( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef stream_EMA( np.ndarray real , int timeperiod)
cdef stream_EXP( np.ndarray real )
cdef stream_FLOOR( np.ndarray real )
cdef stream_HT_DCPERIOD( np.ndarray real )
cdef stream_HT_DCPHASE( np.ndarray real )
cdef stream_HT_PHASOR( np.ndarray real )
cdef stream_HT_SINE( np.ndarray real )
cdef stream_HT_TRENDLINE( np.ndarray real )
cdef stream_HT_TRENDMODE( np.ndarray real )
cdef stream_KAMA( np.ndarray real , int timeperiod)
cdef stream_LINEARREG( np.ndarray real , int timeperiod)
cdef stream_LINEARREG_ANGLE( np.ndarray real , int timeperiod)
cdef stream_LINEARREG_INTERCEPT( np.ndarray real , int timeperiod)
cdef stream_LINEARREG_SLOPE( np.ndarray real , int timeperiod)
cdef stream_LN( np.ndarray real )
cdef stream_LOG10( np.ndarray real )
cdef stream_MA( np.ndarray real , int timeperiod, int matype)
cdef tuple_double3 stream_MACD( np.ndarray real , int fastperiod, int slowperiod, int signalperiod)
cdef stream_MACDEXT( np.ndarray real , int fastperiod, int fastmatype, int slowperiod, int slowmatype, int signalperiod, int signalmatype)
cdef stream_MACDFIX( np.ndarray real , int signalperiod)
cdef stream_MAMA( np.ndarray real , double fastlimit, double slowlimit)
cdef stream_MAVP( np.ndarray real , np.ndarray periods , int minperiod, int maxperiod, int matype)
cdef stream_MAX( np.ndarray real , int timeperiod)
cdef stream_MAXINDEX( np.ndarray real , int timeperiod)
cdef stream_MEDPRICE( np.ndarray high , np.ndarray low )
cdef stream_MFI( np.ndarray high , np.ndarray low , np.ndarray close , np.ndarray volume , int timeperiod)
cdef stream_MIDPOINT( np.ndarray real , int timeperiod)
cdef stream_MIDPRICE( np.ndarray high , np.ndarray low , int timeperiod)
cdef stream_MIN( np.ndarray real , int timeperiod)
cdef stream_MININDEX( np.ndarray real , int timeperiod)
cdef stream_MINMAX( np.ndarray real , int timeperiod)
cdef stream_MINMAXINDEX( np.ndarray real , int timeperiod)
cdef stream_MINUS_DI( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef stream_MINUS_DM( np.ndarray high , np.ndarray low , int timeperiod)
cdef stream_MOM( np.ndarray real , int timeperiod)
cdef stream_MULT( np.ndarray real0 , np.ndarray real1 )
cdef stream_NATR( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef stream_OBV( np.ndarray real , np.ndarray volume )
cdef stream_PLUS_DI( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef stream_PLUS_DM( np.ndarray high , np.ndarray low , int timeperiod)
cdef stream_PPO( np.ndarray real , int fastperiod, int slowperiod, int matype)
cdef stream_ROC( np.ndarray real , int timeperiod)
cdef stream_ROCP( np.ndarray real , int timeperiod)
cdef stream_ROCR( np.ndarray real , int timeperiod)
cdef stream_ROCR100( np.ndarray real , int timeperiod)
cdef stream_RSI( np.ndarray real , int timeperiod)
cdef stream_SAR( np.ndarray high , np.ndarray low , double acceleration, double maximum)
cdef stream_SAREXT( np.ndarray high , np.ndarray low , double startvalue, double offsetonreverse, double accelerationinitlong, double accelerationlong, double accelerationmaxlong, double accelerationinitshort, double accelerationshort, double accelerationmaxshort)
cdef stream_SIN( np.ndarray real )
cdef stream_SINH( np.ndarray real )
cdef stream_SMA( np.ndarray real , int timeperiod)
cdef stream_SQRT( np.ndarray real )
cdef stream_STDDEV( np.ndarray real , int timeperiod, double nbdev)
cdef tuple_double2 stream_STOCH( np.ndarray high , np.ndarray low , np.ndarray close , int fastk_period, int slowk_period, int slowk_matype, int slowd_period, int slowd_matype)
cdef stream_STOCHF( np.ndarray high , np.ndarray low , np.ndarray close , int fastk_period, int fastd_period, int fastd_matype)
cdef stream_STOCHRSI( np.ndarray real , int timeperiod, int fastk_period, int fastd_period, int fastd_matype)
cdef stream_SUB( np.ndarray real0 , np.ndarray real1 )
cdef stream_SUM( np.ndarray real , int timeperiod)
cdef stream_T3( np.ndarray real , int timeperiod, double vfactor)
cdef stream_TAN( np.ndarray real )
cdef stream_TANH( np.ndarray real )
cdef stream_TEMA( np.ndarray real , int timeperiod)
cdef stream_TRANGE( np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_TRIMA( np.ndarray real , int timeperiod)
cdef stream_TRIX( np.ndarray real , int timeperiod)
cdef stream_TSF( np.ndarray real , int timeperiod)
cdef stream_TYPPRICE( np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_ULTOSC( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod1, int timeperiod2, int timeperiod3)
cdef stream_VAR( np.ndarray real , int timeperiod, double nbdev)
cdef stream_WCLPRICE( np.ndarray high , np.ndarray low , np.ndarray close )
cdef stream_WILLR( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef stream_WMA( np.ndarray real , int timeperiod)

# END FUNCTIONS GEN FROM '_ta_lib_stream.pxi'
#############################################


#############################################
# START FUNCTIONS GEN FROM '_numpy_funcs.pxi'

cdef void shift_inplace(double[::1] arr, int num) noexcept nogil
cdef np.ndarray shift(np.ndarray arr, int num)
cdef void replace(double[::1] arr, double orig, double value) noexcept nogil
cdef void ffill(double[::1] arr) noexcept nogil
cdef np.ndarray rolling_sum(numeric_dtype[::1] arr, int window)

# END FUNCTIONS GEN FROM '_numpy_funcs.pxi'
#############################################


#############################################
# START FUNCTIONS GEN FROM '_ta_lib_func.pxi'

cdef np.ndarray make_double_array(np.npy_intp length, int lookback)
cdef np.ndarray make_int_array(np.npy_intp length, int lookback)
cdef ACOS( np.ndarray real )
cdef AD( np.ndarray high , np.ndarray low , np.ndarray close , np.ndarray volume )
cdef ADD( np.ndarray real0 , np.ndarray real1 )
cdef ADOSC( np.ndarray high , np.ndarray low , np.ndarray close , np.ndarray volume , int fastperiod, int slowperiod)
cdef ADX( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef ADXR( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef APO( np.ndarray real , int fastperiod, int slowperiod, int matype)
cdef AROON( np.ndarray high , np.ndarray low , int timeperiod)
cdef AROONOSC( np.ndarray high , np.ndarray low , int timeperiod)
cdef ASIN( np.ndarray real )
cdef ATAN( np.ndarray real )
cdef ATR( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef AVGPRICE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef BBANDS( np.ndarray real , int timeperiod, double nbdevup, double nbdevdn, int matype)
cdef BETA( np.ndarray real0 , np.ndarray real1 , int timeperiod)
cdef BOP( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CCI( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef CDL2CROWS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDL3BLACKCROWS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDL3INSIDE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDL3LINESTRIKE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDL3OUTSIDE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDL3STARSINSOUTH( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDL3WHITESOLDIERS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLABANDONEDBABY( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef CDLADVANCEBLOCK( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLBELTHOLD( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLBREAKAWAY( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLCLOSINGMARUBOZU( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLCONCEALBABYSWALL( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLCOUNTERATTACK( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLDARKCLOUDCOVER( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef CDLDOJI( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLDOJISTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLDRAGONFLYDOJI( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLENGULFING( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLEVENINGDOJISTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef CDLEVENINGSTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef CDLGAPSIDESIDEWHITE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLGRAVESTONEDOJI( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLHAMMER( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLHANGINGMAN( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLHARAMI( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLHARAMICROSS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLHIGHWAVE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLHIKKAKE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLHIKKAKEMOD( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLHOMINGPIGEON( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLIDENTICAL3CROWS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLINNECK( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLINVERTEDHAMMER( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLKICKING( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLKICKINGBYLENGTH( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLLADDERBOTTOM( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLLONGLEGGEDDOJI( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLLONGLINE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLMARUBOZU( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLMATCHINGLOW( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLMATHOLD( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef CDLMORNINGDOJISTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef CDLMORNINGSTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close , double penetration)
cdef CDLONNECK( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLPIERCING( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLRICKSHAWMAN( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLRISEFALL3METHODS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLSEPARATINGLINES( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLSHOOTINGSTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLSHORTLINE( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLSPINNINGTOP( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLSTALLEDPATTERN( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLSTICKSANDWICH( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLTAKURI( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLTASUKIGAP( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLTHRUSTING( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLTRISTAR( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLUNIQUE3RIVER( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLUPSIDEGAP2CROWS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CDLXSIDEGAP3METHODS( np.ndarray open , np.ndarray high , np.ndarray low , np.ndarray close )
cdef CEIL( np.ndarray real )
cdef CMO( np.ndarray real , int timeperiod)
cdef CORREL( np.ndarray real0 , np.ndarray real1 , int timeperiod)
cdef COS( np.ndarray real )
cdef COSH( np.ndarray real )
cdef DEMA( np.ndarray real , int timeperiod)
cdef DIV( np.ndarray real0 , np.ndarray real1 )
cdef DX( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef EMA( np.ndarray real , int timeperiod)
cdef EXP( np.ndarray real )
cdef FLOOR( np.ndarray real )
cdef HT_DCPERIOD( np.ndarray real )
cdef HT_DCPHASE( np.ndarray real )
cdef HT_PHASOR( np.ndarray real )
cdef HT_SINE( np.ndarray real )
cdef HT_TRENDLINE( np.ndarray real )
cdef HT_TRENDMODE( np.ndarray real )
cdef KAMA( np.ndarray real , int timeperiod)
cdef LINEARREG( np.ndarray real , int timeperiod)
cdef LINEARREG_ANGLE( np.ndarray real , int timeperiod)
cdef LINEARREG_INTERCEPT( np.ndarray real , int timeperiod)
cdef LINEARREG_SLOPE( np.ndarray real , int timeperiod)
cdef LN( np.ndarray real )
cdef LOG10( np.ndarray real )
cdef MA( np.ndarray real , int timeperiod, int matype)
cdef MACD( np.ndarray real , int fastperiod, int slowperiod, int signalperiod)
cdef MACDEXT( np.ndarray real , int fastperiod, int fastmatype, int slowperiod, int slowmatype, int signalperiod, int signalmatype)
cdef MACDFIX( np.ndarray real , int signalperiod)
cdef MAMA( np.ndarray real , double fastlimit, double slowlimit)
cdef MAVP( np.ndarray real , np.ndarray periods , int minperiod, int maxperiod, int matype)
cdef MAX( np.ndarray real , int timeperiod)
cdef MAXINDEX( np.ndarray real , int timeperiod)
cdef MEDPRICE( np.ndarray high , np.ndarray low )
cdef MFI( np.ndarray high , np.ndarray low , np.ndarray close , np.ndarray volume , int timeperiod)
cdef MIDPOINT( np.ndarray real , int timeperiod)
cdef MIDPRICE( np.ndarray high , np.ndarray low , int timeperiod)
cdef MIN( np.ndarray real , int timeperiod)
cdef MININDEX( np.ndarray real , int timeperiod)
cdef MINMAX( np.ndarray real , int timeperiod)
cdef MINMAXINDEX( np.ndarray real , int timeperiod)
cdef MINUS_DI( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef MINUS_DM( np.ndarray high , np.ndarray low , int timeperiod)
cdef MOM( np.ndarray real , int timeperiod)
cdef MULT( np.ndarray real0 , np.ndarray real1 )
cdef NATR( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef OBV( np.ndarray real , np.ndarray volume )
cdef PLUS_DI( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef PLUS_DM( np.ndarray high , np.ndarray low , int timeperiod)
cdef PPO( np.ndarray real , int fastperiod, int slowperiod, int matype)
cdef ROC( np.ndarray real , int timeperiod)
cdef ROCP( np.ndarray real , int timeperiod)
cdef ROCR( np.ndarray real , int timeperiod)
cdef ROCR100( np.ndarray real , int timeperiod)
cdef RSI( np.ndarray real , int timeperiod)
cdef SAR( np.ndarray high , np.ndarray low , double acceleration, double maximum)
cdef SAREXT( np.ndarray high , np.ndarray low , double startvalue, double offsetonreverse, double accelerationinitlong, double accelerationlong, double accelerationmaxlong, double accelerationinitshort, double accelerationshort, double accelerationmaxshort)
cdef SIN( np.ndarray real )
cdef SINH( np.ndarray real )
cdef SMA( np.ndarray real , int timeperiod)
cdef SQRT( np.ndarray real )
cdef STDDEV( np.ndarray real , int timeperiod, double nbdev)
cdef STOCH( np.ndarray high , np.ndarray low , np.ndarray close , int fastk_period, int slowk_period, int slowk_matype, int slowd_period, int slowd_matype)
cdef STOCHF( np.ndarray high , np.ndarray low , np.ndarray close , int fastk_period, int fastd_period, int fastd_matype)
cdef STOCHRSI( np.ndarray real , int timeperiod, int fastk_period, int fastd_period, int fastd_matype)
cdef SUB( np.ndarray real0 , np.ndarray real1 )
cdef SUM( np.ndarray real , int timeperiod)
cdef T3( np.ndarray real , int timeperiod, double vfactor)
cdef TAN( np.ndarray real )
cdef TANH( np.ndarray real )
cdef TEMA( np.ndarray real , int timeperiod)
cdef TRANGE( np.ndarray high , np.ndarray low , np.ndarray close )
cdef TRIMA( np.ndarray real , int timeperiod)
cdef TRIX( np.ndarray real , int timeperiod)
cdef TSF( np.ndarray real , int timeperiod)
cdef TYPPRICE( np.ndarray high , np.ndarray low , np.ndarray close )
cdef ULTOSC( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod1, int timeperiod2, int timeperiod3)
cdef VAR( np.ndarray real , int timeperiod, double nbdev)
cdef WCLPRICE( np.ndarray high , np.ndarray low , np.ndarray close )
cdef WILLR( np.ndarray high , np.ndarray low , np.ndarray close , int timeperiod)
cdef WMA( np.ndarray real , int timeperiod)

# END FUNCTIONS GEN FROM '_ta_lib_func.pxi'
#############################################


#############################################
# START FUNCTIONS GEN FROM 'indicators.pyx'

cdef recent_SMA(double[::1] real, int timeperiod, int calc_length)
cdef BIAS(np.ndarray real, int timeperiod)
cdef double stream_BIAS(np.ndarray real, int timeperiod)
cdef recent_BIAS(double[::1] real, int timeperiod, int calc_length)
cdef recent_MACD(double[::1] real, int fastperiod, int slowperiod , int signalperiod, int calc_length )
cdef recent_STOCH(double[::1] high, double[::1] low, double[::1] close, int fastk_period, int slowk_period, int slowk_matype, int slowd_period, int slowd_matype, int calc_length)
cdef KD(np.ndarray high, np.ndarray low, np.ndarray close, int fastk_period, int slowk_period, int slowd_period,)
cdef tuple_double2 stream_KD(np.ndarray high, np.ndarray low, np.ndarray close, int fastk_period, int slowk_period, int slowd_period)
cdef recent_KD(double[::1] high, double[::1] low, double[::1] close, int fastk_period, int slowk_period, int slowd_period, int calc_length)
cdef KDJ(np.ndarray high, np.ndarray low, np.ndarray close, int fastk_period, int slowk_period, int slowd_period)
cdef tuple_double3 stream_KDJ(np.ndarray high, np.ndarray low, np.ndarray close, int fastk_period, int slowk_period, int slowd_period)
cdef recent_KDJ(double[::1] high, double[::1] low, double[::1] close, int fastk_period, int slowk_period, int slowd_period, int calc_length)
cdef SLOW_KD(np.ndarray high, np.ndarray low, np.ndarray close , int fastk_period, int slowkd_period)
cdef tuple_double2 stream_SLOW_KD(double[::1] high, double[::1] low, double[::1] close , int fastk_period, int slowkd_period)
cdef recent_SLOW_KD(double[::1] high, double[::1] low, double[::1] close , int fastk_period, int slowkd_period, int calc_length)
cdef AMPLITUDE(np.ndarray high, np.ndarray low, np.ndarray close, int timeperiod)
cdef double stream_AMPLITUDE(double[::1] high, double[::1] low, double[::1] close, int timeperiod)
cdef recent_AMPLITUDE(np.ndarray high, np.ndarray low, np.ndarray close, int timeperiod, int calc_length)
cdef ZIG(double[::1] real, double perctg)
cdef PERIOD_MAX_BIAS(np.ndarray close, int ma_timeperiod, int period_nums)

# END FUNCTIONS GEN FROM 'indicators.pyx'
#############################################



class TaFormulaError(Exception):
    pass


class DataBackendNotFound(TaFormulaError):
    pass


class PyxSyntaxError(TaFormulaError):
    pass


class DatasListNotMatch(TaFormulaError):
    pass


class TaTimeoutError(TaFormulaError):
    pass



import importlib
import logging
import os
import tempfile
from distutils.sysconfig import get_python_lib
from itertools import chain

import numpy as np
from pyximport import pyximport

from .exceptions import DatasListNotMatch
from .strategy_ast import parse_pyx_file

pyximport.install(build_in_temp=False, setup_args={"include_dirs": np.get_include()})

__all__ = ["Strategy", "get_strategy"]

_temp_pyx_folder = os.path.join(get_python_lib(), "ta_formula_strategies")
os.makedirs(_temp_pyx_folder, mode=0o777, exist_ok=True)
package_init_py = os.path.join(_temp_pyx_folder, "__init__.py")
if not os.path.exists(package_init_py):
    with open(package_init_py, "w") as f:
        f.write("")


def _get_strategy_hash(pyx_file, params, return_fileds):
    return hash(frozenset(chain((pyx_file,), params.items(), return_fileds)))


class Strategy:
    strategies_home_dir = "."  # 当pyx_file传入相对路径时，默认在此目录下查找

    def __init__(
        self,
        pyx_file: str,
        params: dict,
        return_fileds: list,
        _hash: int = 0,
        debug=False,
    ):
        if _hash == 0:
            _hash = _get_strategy_hash(pyx_file, params, return_fileds)
        self._hash = _hash
        if not os.path.isabs(pyx_file):
            pyx_file = os.path.join(self.strategies_home_dir, pyx_file)
        self.pyx_filename = os.path.basename(pyx_file)

        self.pyx_code, pyx_struct = parse_pyx_file(
            pyx_file, params, return_fileds, debug
        )
        self.datas_interface = pyx_struct["datas_interface"]
        self.datas_struct = pyx_struct["datas"]
        self._datas_list = [None] * len(self.datas_interface)

        # compile
        with tempfile.NamedTemporaryFile(
            "w",
            suffix=self.pyx_filename,
            encoding="utf8",
            dir=_temp_pyx_folder,
            delete=False,
        ) as temp_pyx:
            temp_pyx.write(self.pyx_code)
        temp_pyx_bld = temp_pyx.name + "bld"
        with open(temp_pyx_bld, "w", encoding="utf8") as build_file:
            build_file.write(
                """
from distutils.extension import Extension
def make_ext(name, filename):
    return Extension(name=name, sources=[filename],define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")])
"""
            )
        try:
            modulename = os.path.basename(temp_pyx.name)
            old_level = logging.getLogger().level
            # BUG: pyximport will change the level of root logger
            calculator = importlib.import_module(
                "ta_formula_strategies." + modulename.replace(".pyx", "")
            )
            logging.getLogger().setLevel(old_level)
            self.calculate = lambda: calculator.calculate(*self._datas_list)
            self.calculate_x = lambda x: calculator.calculate(*x)
        finally:
            try:
                os.remove(temp_pyx.name)
                os.remove(temp_pyx.name.replace(".pyx", ".c"))
                os.remove(temp_pyx_bld)
            except:
                pass

    def __repr__(self) -> str:
        return self.pyx_code

    def __str__(self) -> str:
        return f"{self.pyx_filename}_{self._hash}"

    def feed_datas(self, datas):
        """数据保存在strategy中，直接调用`calculate`"""
        try:
            for i, (k, v) in enumerate(self.datas_interface):
                self._datas_list[i] = eval(v, {}, {"datas": datas})
        except IndexError as e:
            raise DatasListNotMatch(f"{v} index out of range") from e

    def feed_external_datas(self, datas, datas_list):
        """数据保存在外部的`datas`中，调用`calculate_x`将`datas`传入，用于策略共享"""
        try:
            for i, (k, v) in enumerate(self.datas_interface):
                datas_list[i] = eval(v, {}, {"datas": datas})
        except IndexError as e:
            raise DatasListNotMatch(f"{v} index out of range") from e


strategy_center = {}


def get_strategy(pyx_file: str, params: dict, return_fileds: list):
    # 优先从策略中心取
    _hash = _get_strategy_hash(pyx_file, params, return_fileds)
    strategy = strategy_center.get(_hash, None)
    if strategy is None:
        strategy_center[_hash] = strategy = Strategy(
            pyx_file, params, return_fileds, _hash
        )
    return strategy



from collections import defaultdict

from .datasource import get_backend
from .exceptions import DatasListNotMatch


class RequestParser:
    def __init__(self, request):
        self.request = request
        self.datasources = []

    def get_strategy_params(self):
        pass

    def get_prepare_data_params(self, strategy):
        pass


class DictRequestParser(RequestParser):
    def get_strategy_params(self):
        return (
            self.request["pyx_file"],
            self.request["params"],
            self.request["return_fields"],
        )

    def get_prepare_data_params(self, strategy):
        # datasources: `[["{symbol}@{bid}", ...], ...]`
        datasources = self.request["datasources"]
        datas_struct = self.request.get("datas", strategy.datas_struct)
        for one_unit_sources in datasources:
            if len(datas_struct) != len(one_unit_sources):
                raise DatasListNotMatch(
                    "request params datas and datasources not matched"
                )
            b = []
            for source, intervals in zip(one_unit_sources, datas_struct):
                symbol, bid = source.split("@")
                b.append((get_backend(bid), symbol, intervals))
            self.datasources.append(b)
        # self.datasources: `[[(backend, symbol, intervals), ...], ...]`
        ret = []
        for group_unit_sources in zip(*self.datasources):
            intervals = group_unit_sources[0][2]
            symbols_by_db = defaultdict(set)
            for source in group_unit_sources:
                symbols_by_db[source[0]].add(source[1])
            for db, symbols in symbols_by_db.items():
                ret.append((db, list(symbols), intervals))
        return ret



import logging
import time
from collections import defaultdict

# from line_profiler import LineProfiler

# profile = LineProfiler()

__all__ = ["calculation_center"]
# __all__ = ['calculation_center', 'profile']


class _CalculationCenter:
    def __init__(self, lock=None) -> None:
        self.calc_units = {}
        # 引用计数需要原子操作
        self._lock = lock
        self.ref_counts = defaultdict(int)

    def push(self, strategy, symbol_infos):
        _hash = hash(
            (
                strategy._hash,
                tuple(
                    (db.bid, symbol, *intervals)
                    for db, symbol, intervals in symbol_infos
                ),
            )
        )
        if self._lock is not None:
            self._lock.acquire()
        self.ref_counts[_hash] += 1
        unit = self.calc_units.get(_hash, None)
        if unit is None:
            self.calc_units[_hash] = unit = _CalculateUnit(
                strategy, symbol_infos, _hash
            )
            logging.debug(f"{unit}: New. Ref count {self.ref_counts[_hash]}")
        else:
            logging.debug(f"{unit}: Ref plus. Count {self.ref_counts[_hash]}")
        if self._lock is not None:
            self._lock.release()
        return unit

    def pop(self, unit):
        _hash = unit._hash
        if self._lock is not None:
            self._lock.acquire()
        ref = self.ref_counts[_hash]
        if ref == 0:
            logging.error(f"{unit}: Ref is zero, cannot pop.")
            return
        ref -= 1
        self.ref_counts[_hash] = ref
        if ref == 0:
            logging.debug(f"{unit}: Delete. Ref count {ref}")
            unit = self.calc_units.pop(_hash)
            unit.release()
        else:
            logging.debug(f"{unit}: Ref minus. Count {ref}")
        if self._lock is not None:
            self._lock.release()


class _ConsistentFlag:
    __slots__ = ["states", "_full_state"]

    def __init__(self, full_state) -> None:
        self._full_state = full_state
        self.states = [False] * full_state

    def consistent(self):
        first_state = self.states[0]
        for state in self.states:
            if state != first_state:
                return False
        return True

    def reset(self):
        self.states = [False] * self._full_state


class _Data:
    __slots__ = ["cflag", "data", "index"]

    def __init__(self, cflag, data, index) -> None:
        self.cflag = cflag
        self.data = data
        self.index = index


class _CalculateUnit:
    def __init__(self, strategy, symbol_infos, _hash) -> None:
        self.strategy = strategy
        self.symbol_infos = symbol_infos
        self._hash = _hash
        self._datas_list = [None] * len(strategy.datas_interface)
        self.datas_map = {}
        self._symbol_names = []
        self._cflags = []
        self._waiters = []
        datas = []
        # 可能不同交易所有相同的symbol
        for db, symbol, intervals in symbol_infos:
            self._symbol_names.append((db.bid, symbol))
            _datas = [db._all_datas[symbol][interval] for interval in intervals]
            datas.append(_datas)
            cflag = _ConsistentFlag(len(intervals))
            self._cflags.append(cflag)
            symbols = self.datas_map.setdefault(db.bid, {})
            symbols[symbol] = dict(
                zip(
                    intervals,
                    [_Data(cflag, _data, i) for i, _data in enumerate(_datas)],
                )
            )
            db._calc_units.add(self)
        strategy.feed_external_datas(datas, self._datas_list)

    def release(self):
        for db, _, _ in self.symbol_infos:
            try:
                db._calc_units.remove(self)
            except KeyError:
                pass

    # @profile
    def on_update(self, backend_id, symbol, interval, update_dt):
        # TODO: 如果有多个数据后台多线程运行，这个函数需要考虑线程安全
        try:
            _data = self.datas_map[backend_id][symbol][interval]
        except KeyError:
            return
        # 附加 数据接收时间戳
        start_counter = time.perf_counter_ns()
        data_rec_time = int(time.time() * 1000000)
        _data.cflag.states[_data.index] = True

        # 同一标的的不同频率的数据，要么全部没更新，要么全部更新
        for cflag in self._cflags:
            if not cflag.consistent():
                return

        # 全部重置
        for cflag in self._cflags:
            cflag.reset()
        # 计算入口
        signal = self.strategy.calculate_x(self._datas_list)
        # 附加 计算单元ID
        signal["calc_unit_id"] = self._hash
        signal["data_update_time"] = update_dt
        signal["data_rec_time"] = data_rec_time
        signal["calc_time"] = start_counter
        signal["symbols"] = self._symbol_names
        for waiter in self._waiters:
            waiter.add_result(self._hash, signal)

    def _add_waiter(self, waiter):
        self._waiters.append(waiter)

    def _remove_waiter(self, waiter):
        self._waiters.remove(waiter)

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, value: object) -> bool:
        return id(self) == id(value)

    def __str__(self):
        return f"{self.__class__.__name__}({self.strategy}, {self.symbol_infos}, {self._hash})"



# 这个版本和cython编译的版本性能差不多
# 直接使用indicators模块下的shift
# 不依赖numba
import numba
import numpy as np


@numba.njit
def shift(arr, num, fill_value=np.nan):
    result = np.empty_like(arr)
    if num > 0:
        result[:num] = fill_value
        result[num:] = arr[:-num]
    elif num < 0:
        result[num:] = fill_value
        result[:num] = arr[-num:]
    else:
        result[:] = arr
    return result



import argparse
import ast
import re

from .exceptions import PyxSyntaxError

cache_func_no_return_type_temp = """
    cdef int _{id}_defined = 0
    _{id} = None
    def {id}():
        nonlocal _{id}_defined, _{id}
        if _{id}_defined == 0:
            _{id} = {func}
            _{id}_defined = 1
        return _{id}
"""


def _get_subscript_value(node):
    if not isinstance(node, ast.Subscript):
        return node
    return _get_subscript_value(node.value)


def parse_pyx_file(pyx_file: str, params: dict, return_fileds: list, debug: bool):
    """
    第一行必须是cimport 开头
    必须有datas字段
    必须有ret字段
    只支持赋值语句和函数调用语句
    """
    with open(pyx_file, "r", encoding="utf8") as pyx:
        # cimport cannot parse by ast module
        code = ""
        while "cimport" in (line := pyx.readline()).strip():
            code += "\n"
        code += line
        code += pyx.read()

    try:
        st = ast.parse(code)
    except SyntaxError as e:
        raise PyxSyntaxError(f"{e.msg}, line {e.lineno}, {e.offset}") from e

    pyx_struct = {}

    temp_pyx_struct = {}
    for node in ast.iter_child_nodes(st):
        if not isinstance(node, (ast.Assign, ast.AnnAssign)):
            raise PyxSyntaxError(
                f"is not assignment, line {value.lineno}, {value.col_offset}"
            )
        if "targets" in node._fields:
            # ann = None
            target = node.targets[0]
        elif "target" in node._fields:
            # ann = ast.unparse(node.annotation)
            target = node.target
        if "id" not in target._fields:
            raise PyxSyntaxError(
                f"tuple unpacking not supported, line {value.lineno}, {value.col_offset}"
            )
        temp_pyx_struct[target.id] = node.value
        # temp_pyx_struct[target.id] = (ann, node.value)

    datas_struct = temp_pyx_struct.pop("datas", None)
    if not datas_struct:
        raise PyxSyntaxError(f'"datas" variable cannot be found.')
    pyx_struct["datas"] = ast.literal_eval(datas_struct)

    ret = temp_pyx_struct.pop("ret", None)
    if not ret:
        raise PyxSyntaxError(f'"ret" variable cannot be found.')
    # 选择性返回结果
    if return_fileds:
        ret_keys = []
        ret_values = []
        for ret_key, ret_value in zip(ret.keys, ret.values):
            for field in return_fileds:
                if field in ret_key.value:
                    ret_keys.append(ret_key)
                    ret_values.append(ret_value)
                    break
        ret = ast.Dict(ret_keys, ret_values)
    else:
        ret_keys = ret.keys
        ret_values = ret.values
    pyx_struct["ret"] = ast.unparse(ret)

    pyx_struct["constant_params"] = {}
    pyx_struct["datas_interface"] = []
    pyx_struct["indicators"] = {}
    for var_id, value in temp_pyx_struct.items():
        if isinstance(value, ast.Constant):
            pyx_struct["constant_params"][var_id] = value.value
        else:
            _value = _get_subscript_value(value)
            if isinstance(_value, ast.Name) and _value.id == "datas":
                # datas subscript
                pyx_struct["datas_interface"].append((var_id, ast.unparse(value)))
            elif isinstance(_value, ast.Call):
                pyx_struct["indicators"][var_id] = ast.unparse(value)
            else:
                raise PyxSyntaxError(
                    f"not support this defines, line {value.lineno}, {value.col_offset}"
                )

    # 更新params
    constant_params = pyx_struct["constant_params"]
    for k, v in params.items():
        if k in constant_params:
            constant_params[k] = v

    # 替换参数
    def replace1(match):
        return str(constant_params[match.group(0)])

    constant_params_replace_regex = "|".join(
        r"\b{}\b".format(k) for k in constant_params.keys()
    )
    # 1. 返回中的参数需要替换
    pyx_struct["ret"] = re.sub(
        constant_params_replace_regex, replace1, pyx_struct["ret"]
    )
    # 2. 函数中的参数需要替换
    for _id, func in pyx_struct["indicators"].items():
        pyx_struct["indicators"][_id] = re.sub(
            constant_params_replace_regex, replace1, func
        )
    # 替换函数
    indicator_ids = pyx_struct["indicators"].keys()

    def replace2(match):
        return f"{match.group(0)}()"

    pyx_struct["ret"] = re.sub(
        "|".join(r"\b{}\b".format(k) for k in indicator_ids),
        replace2,
        pyx_struct["ret"],
    )

    ret_key_strs = [k.value for k in ret_keys]
    data_params = [field for field, _ in pyx_struct["datas_interface"]]
    if debug:
        debug_str = str.join(
            ", ", [f"'debug_{field}': {field}" for field in data_params]
        )
        ret_key_strs += [f"debug_{field}" for field in data_params]
    cache_funcs = []
    for _id, func in pyx_struct["indicators"].items():
        cf = cache_func_no_return_type_temp.format(id=_id, func=func)
        cache_funcs.append(cf)
        if debug:
            ret_key_strs.append(f"debug_{_id}")
            debug_str += ", "
            debug_str += f"'debug_{_id}': {_id}()"
    if debug:
        pyx_struct["ret"] = pyx_struct["ret"][:-1] + ", " + debug_str + "}"
    ret_str = _dict_formatter(pyx_struct["ret"], ret_key_strs)
    code = f"""
# THIS IS AUTO GENERATED FILE, DO NOT MODIFY THIS FILE

#cython: language_level=3str
cimport ta_formula.indicators as ta


def calculate({str.join(', ', [field for field in data_params])}):
    {str.join('', cache_funcs)}

    return {ret_str}
    """
    return code, pyx_struct


key_value_pair = re.compile(r"[\'\"](\w+)[\'\"]\s*:(.*)")


def _dict_formatter(dict_str: str, keys):
    # 选择性返回结果
    s = [dict_str.index(k) - 1 for k in keys]
    rebuild_dict = "{\n"
    for i, j in zip(s, s[1:] + [-1]):
        m = re.match(key_value_pair, dict_str[i:j])
        k, v = m.group(1), m.group(2)
        rebuild_dict += f"        '{k}':{v}\n"
    rebuild_dict += "    }\n"
    return rebuild_dict


def tacompile_entry():
    parser = argparse.ArgumentParser(description="将pyx策略模板文件编译成策略文件")
    parser.add_argument("pyx_file", help="pyx策略模板")
    parser.add_argument("-d", "--debug", action="store_true", help="是否输出debug数据")
    args = parser.parse_args()
    code, _ = parse_pyx_file(args.pyx_file, {}, [], args.debug)
    output_file = args.pyx_file.replace(".pyx", "_out.pyx")
    with open(output_file, "w", encoding="utf8") as f:
        f.write(code)


if __name__ == "__main__":
    code, pyx_struct = parse_pyx_file("test_strategy.pyx", {}, ["open", "close"], False)
    print(code)
